#rfd-overlay {
  z-index: 1000;
  position: fixed;
  top: 0;
  left: 0;
  width: 100%;
  height: 100%;
  background-color: rgba(0, 0, 0, 0.5);

  animation: init 0.5s;

  display: flex;
  justify-content: center;
  align-items: center;
}
#rfd-overlay.hidden {
    display: none;
}
#rfd-card {
  padding: 20px;
  border-radius: 5px;

  box-shadow: 0 24px 38px 3px rgba(0, 0, 0, 0.14),
    0 9px 46px 8px rgba(0, 0, 0, 0.12), 0 11px 15px -7px rgba(0, 0, 0, 0.2);

  color-scheme: light dark;
  background-color: canvas;
  color: canvastext;
}
#rfd-title {
  line-height: 1.6;
}
#rfd-input,#rfd-output {
  text-align: center;
}
.rfd-button {
  display: block;
  margin-top: 5px;
  width: 100%;
}

@keyframes init {
  0% {
    opacity: 0;
  }
  100% {
    opacity: 1;
  }
}
